/*= R&S OSP Open Switch and Control Platform ================================
 * LabWindows/CVI 8.0 VXIpnp Instrument Driver
 * Original Release: July 2008
 * By: Lukas Hynek
 *
 * Should you have any technical questions please contact the hotline of
 * Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 * e-mail: CustomerSupport@rsd.rohde-schwarz.com
 *
 * Modification History:
 *   see file rsosp_vxi.chm
 *===========================================================================*/
#ifdef _WIN32
    #ifdef _CVI_
        #include <userint.h>
    #else
        #include <windows.h>
    #endif
#endif
#include <math.h>
#include <visa.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include "rsosp.h"
#include "rsosp_utility.h"
#include <limits.h>
#include <locale.h>  

/*= Macros ==================================================================*/

#define CHECKVIINT32(io, arg, min, max) if ( rsosp_invalidViInt32Range (io, arg,min,max)) return rsosp_ParameterError(io,#arg)

#define SETPARAMINT32(io,cmd,value,min,max) rsosp_setViInt32 (io,cmd,value,min,max,#value)
#define GETPARAMINT32(io,cmd,value) rsosp_getViInt32 (io,cmd,value,#value)
#define GETPARAMINT32ARRAY(io,cmd,value,count) rsosp_getViInt32Array (io,cmd,value,count,#value)
#define GETPARAMUINT32(io,cmd,value) rsosp_getViUInt32 (io,cmd,value,#value)

#define SETPARAMSTRING(io,cmd,value) rsosp_setViString (io,cmd,value,#value)
#define GETPARAMSTRING(io,cmd,value,buf_size) rsosp_getViString (io,cmd,value,buf_size,#value)

#define SETPARAMSTRINGQUOTED(io,cmd,quotes,value) rsosp_setViStringQuoted (io,cmd,quotes,value,#value)
#define GETPARAMSTRINGQUOTED(io,cmd,quotes,buf_size,value) rsosp_getViStringQuoted (io,cmd,quotes,buf_size,value,#value)

#define SETPARAMBOOL(io,cmd,value) rsosp_setViBoolean (io,cmd,value,#value)
#define GETPARAMBOOL(io,cmd,value) rsosp_getViBoolean (io,cmd,value,#value)

#define CHECKERR_CU(fCal) if (rsosp_status = (fCal), rsosp_status < VI_SUCCESS) \
    return rsosp_initCleanUp (rmSession, instrSession, rsosp_status); else

/*= DEFINES =================================================================*/
#define RSOSP_REVISION          "Rev 1.0.2, 09/2019, CVI 2019" /* Instrument driver revision */
#define BUFFER_SIZE             2048L    /* File I/O buffer size */

#define RSOSP_DEFAULT_TIMEOUT   10000   /* Default Timeout value in ms (10 sec) */

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT COMMAND ARRAYS =====================================*/
/*****************************************************************************/

/*****************************************************************************/
/*= UTILITY ROUTINE DECLARATIONS (Non-Exportable Functions) =================*/
/*****************************************************************************/
ViStatus rsosp_initCleanUp (ViSession openRMSession, ViPSession openInstrSession,
                             ViStatus currentStatus);
ViStatus rsosp_defaultInstrSetup (ViSession openInstrSession);

/*****************************************************************************
 * Function: Initialize
 * Purpose:  This function performs the following initialization actions:
 *           
 *           - Opens a session to the Default Resource Manager resource and a 
 *           session to the specified device using the interface and address 
 *           specified in the Resource_Name control.
 *           
 *           - Performs an identification query on the Instrument.
 *           
 *           - Resets the instrument to a known state.
 *           
 *           - Sends initialization commands to the instrument that set any 
 *           necessary programmatic variables such as Headers Off, Short 
 *           Command form, and Data Transfer Binary to the state necessary for 
 *           the operation of the instrument driver.
 *           
 *           - Returns an Instrument Handle which is used to differentiate 
 *           between different sessions of this instrument driver.
 *           
 *           - Each time this function is invoked a Unique Session is opened.  
 *           It is possible to have more than one session open for the same 
 *           resource.
 *           
 *****************************************************************************/
ViStatus _VI_FUNC rsosp_init (
    ViRsrc      resourceName, 
    ViBoolean   idQuery,
    ViBoolean   resetDevice, 
    ViSession   *instrSession
)
{
    ViStatus            rsosp_status = VI_SUCCESS;
    ViSession           rmSession = 0;
    ViChar              buffer[BUFFER_SIZE] = "";
    rsosp_instrRange    instrPtr = NULL;

    /* --- Set standard "C" locale --- */
    {
       char *locale;

       locale = setlocale(LC_ALL, "C");
    }
    
    /*- Open instrument session ---------------------------------------------*/
    CHECKERR( viOpenDefaultRM (&rmSession));

    if ((rsosp_status = viOpen (rmSession, resourceName, VI_LOAD_CONFIG,
                                 VI_NULL, instrSession)) < 0)
    {
        (void)viClose (rmSession);
        return rsosp_status;
    }

    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */
    if (instrPtr == NULL)
        instrPtr = calloc (1, sizeof (struct rsosp_statusDataRanges));
        

    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_USER_DATA, (ViAttrState)instrPtr));

    //With remote control via Ethernet, remote control is not automatically
    //set by means of a command. The instrument must be explicitly set to
    //the REMOTE state, e.g. by sending the intf command &GTR
    CHECKERR (viPrintf (*instrSession, "&GTR\n"));

    /*- Configure VISA Formatted I/O ----------------------------------------*/
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_TMO_VALUE,
                                 RSOSP_DEFAULT_TIMEOUT));
    CHECKERR( viSetBuf (*instrSession, (ViUInt16) (VI_READ_BUF|VI_WRITE_BUF), 4000));
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_WR_BUF_OPER_MODE,
                                 VI_FLUSH_ON_ACCESS));
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_RD_BUF_OPER_MODE,
                                 VI_FLUSH_ON_ACCESS));

    /* -------------------------------------------------------------------- */
    /* Drawback of Agilent VISA is patched by adding extra termination      */
    /* character "\n" to each I/O operation                                 */
    /*                                                                      */
    /* VI_ATTR_SEND_END_EN will stay with this attribute set to VI_TRUE     */
    /* -------------------------------------------------------------------- */

    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_SEND_END_EN, VI_TRUE));

    // read/reset ESR which may still indicate device dependent error from 
    // previous handling!
    (void) rsosp_checkStatus(*instrSession);
    
    /*- Identification Query ------------------------------------------------*/
    if (idQuery == VI_TRUE)
    {
        CHECKERR( viQueryf (*instrSession, "*IDN?\n", "%*[^,],%[^,],%*[\r\n]",
                               buffer));

        /* Check if instrument is supported or not */
        RemoveSurroundingWhiteSpace(buffer);    
        if (strncmp (buffer, "OSP", 3) != 0)
            return rsosp_initCleanUp (rmSession, instrSession, VI_ERROR_FAIL_ID_QUERY);
    }

    /*- Reset instrument ----------------------------------------------------*/
    if (resetDevice == VI_TRUE)
        CHECKERR( rsosp_reset (*instrSession));
    else  /*- Send Default Instrument Setup ---------------------------------*/
        CHECKERR( rsosp_defaultInstrSetup (*instrSession));

    return rsosp_status;
}


/*============================================================================
 * Function: rsosp_SetListOfRelayConditions
 * Purpose:  This function allows specific individual relays and I/O channels
 *           to be set.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SetListOfRelayConditions (
    ViSession   instrSession,
    ViString    channelList
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKNULL (channelList);

    pbuffer = malloc ((strlen (channelList) + 14)*sizeof(ViChar));
    
    sprintf (pbuffer, "ROUT:CLOS (@%s)", channelList);
    
    rsosp_status = rsosp_sendCmd (instrSession, pbuffer);   
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;    
}

/*============================================================================
 * Function: rsosp_GetListOfRelayConditions
 * Purpose:  This query function allows the condition of individual relays
 *           and I/O channels to be queried. The instrument returns a 1 or 0
 *           for each relay or channel in the list, in the same order as the
 *           list is specified. A response of 1 means the relay or the
 *           channel is in the condition that is asked for in the Channel
 *           List. A response of 0 means the channel is not in the condition
 *           that is asked for in the Channel List.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetListOfRelayConditions (
    ViSession   instrSession,
    ViString    channelList,
    ViInt32     arraySize,
    ViUInt32    _VI_FAR conditions[]
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKNULL (channelList);
    CHECKNULL (conditions);

    pbuffer = malloc ((strlen (channelList) + 15)*sizeof(ViChar));
    
    sprintf (pbuffer, "ROUT:CLOS? (@%s)", channelList);
    
    rsosp_status = rsosp_getViUInt32Array (instrSession, 
                                           pbuffer,
                                           conditions,
                                           arraySize,
                                           "Conditions");    
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;
}

/*============================================================================
 * Function: rsosp_SetRelayCondition
 * Purpose:  This function allows specific individual relay and I/O channel
 *           to be set.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SetRelayCondition (
    ViSession   instrSession,
    ViInt32     instrumentName,
    ViInt32     moduleName,
    ViUInt32    condition,
    ViUInt32    relayNumber
)
{
    ViChar      buffer[BUFFER_SIZE];
    
    CHECKVIINT32 (instrSession, instrumentName, 1, 9);
    CHECKVIINT32 (instrSession, moduleName, 1, 3);
    
    sprintf (buffer, "ROUT:CLOS (@F0%dA1%d(%02ld%02ld))", instrumentName, moduleName, condition, relayNumber);
    
    return rsosp_sendCmd (instrSession, buffer);
}

/*============================================================================
 * Function: rsosp_GetRelayCondition
 * Purpose:  This query function allows the condition of individual relay or
 *           I/O channel to be queried. The instrument returns a 1 or 0 for
 *           selected relay or channel. A response of 1 means the relay or
 *           the channel is in the condition that is asked for in the
 *           Condition. A response of 0 means the channel is not in the
 *           condition that is asked for in the Condition.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetRelayCondition (
    ViSession   instrSession,
    ViInt32     instrumentName,
    ViInt32     moduleName,
    ViUInt32    condition,
    ViUInt32    relayNumber,
    ViUInt32    *queryCondition
)
{
    ViChar      buffer[BUFFER_SIZE];
    
    CHECKVIINT32 (instrSession, instrumentName, 1, 9);
    CHECKVIINT32 (instrSession, moduleName, 1, 3);
    CHECKNULL (queryCondition);
    
    sprintf (buffer, "ROUT:CLOS? (@F0%dA1%d(%02ld%02ld))", instrumentName, moduleName, condition, relayNumber);
    
    return GETPARAMUINT32 (instrSession, 
                              buffer, 
                              queryCondition);
}

/*============================================================================
 * Function: rsosp_SetRelayDelay
 * Purpose:  At the moment this function applies only to the four power
 *           relays located on module B104. It allows to set a time delay for
 *           relay operations on module B104. After a "Set List Of Relay
 *           Conditions" or "Set Relay Condition" that modifies the condition
 *           of a relay on B104, the instrument only accepts further
 *           functions after this period of time. This makes sure that the
 *           condition change process of a power relay has completed before
 *           further functions are executed.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SetRelayDelay (
    ViSession   instrSession,
    ViInt32     delayTime, 
    ViString    channelList
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;

    CHECKERR( rsosp_checkOption(instrSession, channelList, B104));
    
    CHECKVIINT32 (instrSession, delayTime, 0, 255);
    CHECKNULL (channelList);

    pbuffer = malloc ((strlen (channelList) + 32)*sizeof(ViChar));
    
    sprintf (pbuffer, "CONFIGURE:RELAY:DELAY %ld,(@%s)", delayTime, channelList);
    
    rsosp_status = rsosp_sendCmd (instrSession, pbuffer);   
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;    
}

/*============================================================================
 * Function: rsosp_GetRelayDelay
 * Purpose:  At the moment this function applies only to the four power
 *           relays located on module B104. It allows to get a time delay for
 *           relay operations on module B104.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetRelayDelay (
    ViSession   instrSession,
    ViString    channelList, 
    ViInt32     arraySize,
    ViInt32     _VI_FAR delayTime[]
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKERR( rsosp_checkOption(instrSession, channelList, B104));
    
    CHECKNULL (channelList);
    CHECKNULL (delayTime);
    
    pbuffer = malloc ((strlen (channelList) + 32)*sizeof(ViChar));
    
    sprintf (pbuffer, "CONFIGURE:RELAY:DELAY? (@%s)", channelList);
    
    rsosp_status = GETPARAMINT32ARRAY (instrSession, 
                                          pbuffer,
                                          delayTime,
                                          arraySize); 
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;    
}

/*============================================================================
 * Function: rsosp_SetOSPName
 * Purpose:  This function assigns an instrument name to a number
 *           representing a R&S OSP instrument address. The instrument name
 *           must have the format F00 up to F09. The instrument address
 *           is a number, whereas the first digit represents the type of the
 *           R&S OSP instrument and the following digits represent the serial
 *           number of the instrument. Once an instrument is assigned to an
 *           instrument name, the relays and I/O channels of this instrument
 *           can be controlled by channel lists built up from the instrument
 *           name, module name, relay number and relay condition. For more
 *           information on how a channel list is built up, refer to chapter
 *           Naming Conventions in operation manual.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SetOSPName (
    ViSession   instrSession,
    ViString    instrumentName,
    ViString    instrumentAddress
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKNULL (instrumentName);
    CHECKNULL (instrumentAddress);
    
    pbuffer = malloc ((strlen (instrumentName) + strlen (instrumentAddress) + 18)*sizeof(ViChar));
    
    sprintf (pbuffer, ":ROUT:MOD:DEF %s,%s", instrumentName, instrumentAddress);
    
    rsosp_status = rsosp_sendCmd (instrSession, pbuffer);   
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;    
}

/*============================================================================
 * Function: rsosp_GetOSPAddress
 * Purpose:  This query function returns the instrument address that has been
 *           previously assigned to the instrument name.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetOSPAddress (
    ViSession   instrSession,
    ViString    instrumentName,
    ViInt32     bufferSize,
    ViChar      _VI_FAR instrumentAddress[]
)
{
    ViChar buffer[BUFFER_SIZE];
    
    CHECKNULL (instrumentName);
    CHECKNULL (instrumentAddress); 
    
    sprintf (buffer, "ROUT:MOD:DEF? %s", instrumentName);
    
    return GETPARAMSTRING (instrSession,
                              buffer,
                              instrumentAddress,
                              bufferSize);
}

/*============================================================================
 * Function: rsosp_DeleteOSP
 * Purpose:  The previously defined Module Name will be deleted from the
 *           instruments internal memory and from the compact flash.
 *           Afterwards the relays and I/O channels of this instrument cannot
 *           be controlled any longer.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_DeleteOSP (
    ViSession   instrSession,
    ViString    instrumentName
)
{
    return SETPARAMSTRING (instrSession,
                            "ROUT:MOD:DEL",
                            instrumentName);
}


/*============================================================================
 * Function: rsosp_DeleteAllOSPs
 * Purpose:  All previously defined Module Names will be deleted from the
 *           instruments internal memory and from the compact flash.
 *           Afterwards the relays and I/O channels of all instruments that
 *           built up a system cannot be controlled any longer.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_DeleteAllOSPs (ViSession instrSession)
{
    return rsosp_sendCmd (instrSession, "ROUT:MOD:DEL:ALL");    
}

/*============================================================================
 * Function: rsosp_UpdateInternalData
 * Purpose:  This function causes an update of all internal held data of the
 *           system of one or several R&S OSP instruments connected via CAN
 *           Bus. The internal held data comprises data like serial number,
 *           part number, hardware code etc. of all modules of all R&S OSP
 *           instruments that were previously announced to the system with
 *           function rsosp_SetModuleName. The data is read from the flash
 *           EEPROM of all modules.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_UpdateInternalData (ViSession instrSession)
{
    return rsosp_sendCmd (instrSession, "ROUT:MOD:INIT");    
}

/*============================================================================
 * Function: rsosp_OSPCatalog
 * Purpose:  This query function returns a list of all relay or I/O channel
 *           modules of all R&S OSP instruments that were previously
 *           announced to the system with function rsosp_SetModuleName.
 *           Besides the internal module name ( e.g. F01A11 ), also the type
 *           of the module is returned ( e.g. OSP-B101 ).
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_OSPCatalog (
    ViSession   instrSession,
    ViInt32     arraySize, 
    ViChar      _VI_FAR catalogList[]
)
{
    return rsosp_getCatalog (instrSession,
                             "ROUT:MOD:CAT?",
                             (ViUInt32) arraySize,
                             catalogList,
                             "catalogList");    
}

/*============================================================================
 * Function: rsosp_SetPathDefinition
 * Purpose:  This function assigns Path Name as a user-specified way of
 *           referring to Channel List. This function allows the user to
 *           define a list of relays and I/O channels to be set to a specific
 *           condition with the use of a Path Name.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SetPathDefinition (
    ViSession   instrSession, 
    ViString    pathName,
    ViString    channelList
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKNULL (pathName);
    CHECKNULL (channelList);
    
    pbuffer = malloc ((strlen (pathName) + strlen (channelList) + 17)*sizeof(ViChar));
    
    sprintf (pbuffer, "ROUT:PATH \"%s\",(@%s)", pathName, channelList);
    
    rsosp_status = rsosp_sendCmd (instrSession, pbuffer);   
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;    
}

/*============================================================================
 * Function: rsosp_GetPathDefinition
 * Purpose:  This query function returns information about what conditions
 *           are defined for all relays and I/O channels that appear in the
 *           Channel List associated with Path Name.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetPathDefinition (
    ViSession   instrSession, 
    ViString    pathName,
    ViInt32     arraySize, 
    ViChar      _VI_FAR channelList[]
)
{
    ViChar      buffer[BUFFER_SIZE];

    CHECKNULL (pathName);
    CHECKNULL (channelList);
    
    sprintf (buffer, "ROUT:PATH? \"%s\"", pathName);    

    return rsosp_getCatalog (instrSession,
                             buffer,
                             (ViUInt32) arraySize,
                             channelList,
                             "channelList");
}

/*============================================================================
 * Function: rsosp_SwitchPath
 * Purpose:  This function connect specific path.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SwitchPath (
    ViSession   instrSession,
    ViString    pathName
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKNULL (pathName);

    pbuffer = malloc ((strlen (pathName) + 14)*sizeof(ViChar));
    
    sprintf (pbuffer, "ROUT:CLOS \"%s\"", pathName);
    
    rsosp_status = rsosp_sendCmd (instrSession, pbuffer);   
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;    
}

/*============================================================================
 * Function: rsosp_DeletePathDefinition
 * Purpose:  The previously defined Path Name will be deleted from the
 *           instruments internal memory.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_DeletePathDefinition (
    ViSession   instrSession,
    ViString    pathName
)
{
    return SETPARAMSTRINGQUOTED (instrSession,
                            "ROUT:PATH:DEL",
                            "\"",
                            pathName);    
}

/*============================================================================
 * Function: rsosp_DeleteAllPathDefinitions
 * Purpose:  All previously defined path names will be deleted from the
 *           instruments internal memory.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_DeleteAllPathDefinitions (ViSession instrSession)
{
    return rsosp_sendCmd (instrSession, "ROUT:PATH:DEL:ALL");    
}

/*============================================================================
 * Function: rsosp_PathCatalog
 * Purpose:  This query function returns a list of all currently defined path
 *           names that are stored in the internal memory of the instrument.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_PathCatalog (
    ViSession   instrSession,
    ViInt32     arraySize, 
    ViChar      _VI_FAR catalogList[]
)
{
    return rsosp_getCatalog (instrSession,
                             "ROUT:PATH:CAT?",
                             (ViUInt32) arraySize,
                             catalogList,
                             "catalogList");    
}

/*============================================================================
 * Function: rsosp_ReadOperationCycles
 * Purpose:  Acquires the number of operation cycles of each relay ( and even
 *           of I/O channels ) that appears in the Channel List. The
 *           instrument returns a number for each relay or channel in the
 *           list, in the same order as the list is specified. For example a
 *           response of 100 means the relay has been closed a hundred times
 *           and opened again in its life time. The number of operation
 *           cycles of each relay is stored durable in the flash EEPROM of
 *           the module the relay is part of. This is done after every hour
 *           of operation time of a R&S OSP ( only if the counter has changed).
 *           This is also done when the instrument receives the query. To
 *           make sure that no operation cycles get lost this READ query
 *           function should be sent before a remote control session is quit.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_ReadOperationCycles (
    ViSession   instrSession,
    ViString    channelList,
    ViInt32     arraySize,
    ViInt32     _VI_FAR operationCycles[]
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;
    
    CHECKNULL (channelList);
    CHECKNULL (operationCycles);     

    pbuffer = malloc ((strlen (channelList) + 32)*sizeof(ViChar));
    
    sprintf (pbuffer, "READ:RELAY:OPERATIONS? (@%s)", channelList);
    
    rsosp_status = GETPARAMINT32ARRAY (instrSession, 
                                          pbuffer,
                                          operationCycles,
                                          arraySize);
    
    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;   
}

/*============================================================================
 * Function: rsosp_ReadInterlockState
 * Purpose:  Acquires the interlock state of a relay module. At the moment
 *           this function applies only to module B104 and is ignored if
 *           Module Name specifies any other module. For more information
 *           about the meaning of interlock, please refer to the description
 *           of module B104.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_ReadInterlockState (
    ViSession   instrSession,
    ViString    moduleName,
    ViBoolean   *interlockState
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer;    
    
    CHECKNULL (moduleName);
    CHECKNULL (interlockState);
    
    CHECKERR( rsosp_checkOption(instrSession, moduleName, B104));
    
    pbuffer = malloc ((strlen (moduleName) + 26)*sizeof(ViChar));
    
    sprintf (pbuffer, "READ:MODULE:INTERLOCK? \"%s\"", moduleName);
    
    rsosp_status =  GETPARAMBOOL (instrSession,
                                        pbuffer,
                                        interlockState);   

    if (pbuffer)
        free (pbuffer);
    
    return rsosp_status;
}

/*============================================================================
 * Function: rsosp_ReadInputChannelsCondition
 * Purpose:  Acquires the condition of all input channels of an I/O module.
 *           Up to sixteen input channels can be queried. I/O module B103 has
 *           sixteen input channels, the I/O section of module B104 has four
 *           input channels. The returned result is an integer value that
 *           represents the condition of all input channels of the module.
 *           This integer value is in the range of 0 ( all channel are
 *           logical 0-LOW ) to 65535 ( all channel are logical 1-HIGH).
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_ReadInputChannelsCondition (
    ViSession   instrSession,
    ViString    moduleName,
    ViUInt32    *channelsCondition
)
{
    ViChar      buffer[BUFFER_SIZE];
    ViStatus    rsosp_status = VI_SUCCESS;  
    ViStatus    rsosp_B103;
    ViStatus    rsosp_B104; 
    ViStatus    rsosp_B105; 
    ViStatus    rsosp_B115; 
    
    CHECKNULL (moduleName);
	CHECKNULL(channelsCondition);
    
    // check module B103
    rsosp_B103 = rsosp_checkOption(instrSession, moduleName, B103);
    // check module B104    
    rsosp_B104 = rsosp_checkOption(instrSession, moduleName, B104); 
    // check module B105    
    rsosp_B105 = rsosp_checkOption(instrSession, moduleName, B105);
    // check module B115    
    rsosp_B115 = rsosp_checkOption(instrSession, moduleName, B115); 
    
    if (rsosp_B103 != VI_SUCCESS && rsosp_B104 != VI_SUCCESS && rsosp_B105 != VI_SUCCESS &&
        rsosp_B115 != VI_SUCCESS)
        CHECKERR(RSOSP_ERROR_INVALID_CONFIGURATION);
    
    if (rsosp_B104 != VI_SUCCESS)
    {
        sprintf (buffer, "READ:IO:IN16? %s", moduleName);       
    }
    else
    {
        sprintf (buffer, "READ:IO:IN4? %s", moduleName);
    }

    return GETPARAMUINT32 (instrSession,
                              buffer,
                              channelsCondition);
}

/*============================================================================
 * Function: rsosp_InternalSelfTest
 * Purpose:  Starts an instrument internal self test routine and returns
 *           information about any error that is detected.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_InternalSelfTest (
    ViSession   instrSession,
    ViInt32     bufferSize,
    ViChar      _VI_FAR selfTestResults[]
)
{
    return GETPARAMSTRING (instrSession,
                              "DIAG:SERV:STES:RES?",
                              selfTestResults,
                              bufferSize);    
}

/*============================================================================
 * Function: rsosp_HardwareInformation
 * Purpose:  Returns information about all components ( mainboards and
 *           modules ) that are part of the complete system of one or several
 *           R&S OSP instruments.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_HardwareInformation (
    ViSession   instrSession,
    ViInt32     bufferSize,
    ViChar      _VI_FAR hwInformation[]
)
{
    return GETPARAMSTRING (instrSession,
                              "DIAG:SERV:HWIN?",
                              hwInformation,
                              bufferSize);    
}

/*============================================================================
 * Function: rsosp_MMemoryCatalog
 * Purpose:  Returns all files that are stored on the compact flash memory.
 *           Each of these files is comprising a set of path configurations.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_MMemoryCatalog (
    ViSession   instrSession,
    ViInt32     arraySize, 
    ViChar      _VI_FAR catalogList[]
)
{
    return rsosp_getCatalog (instrSession,
                             "MMEM:CAT?",
                             (ViUInt32) arraySize,
                             catalogList,
                             "catalogList");
}

/*============================================================================
 * Function: rsosp_MMemoryDelete
 * Purpose:  Removes a file from the compact flash memory.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_MMemoryDelete (
    ViSession   instrSession,
    ViString    fileName
)
{
    return SETPARAMSTRINGQUOTED (instrSession,
                              "MMEM:DEL",
                              "'",
                              fileName);
}

/*============================================================================
 * Function: rsosp_MMemoryLoad
 * Purpose:  Loads a file from the compact flash memory to the instruments
 *           internal memory. The loaded file comprises a set of path
 *           configurations.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_MMemoryLoad (
    ViSession   instrSession,
    ViInt32     instrumentSetting,
    ViString    fileName
)
{
    ViChar buffer[BUFFER_SIZE];

    CHECKNULL (fileName);

    sprintf (buffer, "MMEM:LOAD:STAT %ld,'%s'", instrumentSetting, fileName);

    return rsosp_sendCmd (instrSession, buffer);    
}

/*============================================================================
 * Function: rsosp_MMemoryStore
 * Purpose:  Stores a file durable on the compact flash memory of a R&S OSP
 *           by transfering it from the instruments internal memory. The
 *           stored file comprises the actual defined set of path
 *           configurations ( previously defined with functin rsops_SetPath).
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_MMemoryStore (
    ViSession   instrSession,
    ViInt32     instrumentSetting,
    ViString    fileName
)
{
    ViChar buffer[BUFFER_SIZE];

    CHECKNULL (fileName);

    sprintf (buffer, "MMEM:STOR:STAT %ld,'%s'", instrumentSetting, fileName);

    return rsosp_sendCmd (instrSession, buffer);    
}

/*============================================================================
 * Function: rsosp_SetSystemKeyboardLock
 * Purpose:  This function (Keyboard LOCk) locks or unlocks the local 
 *           controls of the instrument. This includes the front panel keys, 
 *           the keyboard, or other local interfaces. This function only 
 *           applies for R&S OSP type 130 which comes with a display and 
 *           panel keys.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SetSystemKeyboardLock (
    ViSession instrSession,
    ViBoolean keyboardLock
)
{
    return rsosp_setViBoolean (instrSession,
                               "SYST:KLOC",
                               keyboardLock);
}

/*============================================================================
 * Function: rsosp_GetSystemKeyboardLock
 * Purpose:  This function returns if the front panel keyboard of the R&S
 *           Signal Generator including the LOCAL key is disabled or enabled.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetSystemKeyboardLock (
    ViSession instrSession,
    ViBoolean *keyboardLock
)
{
    return GETPARAMBOOL (instrSession,
                               "SYST:KLOC?",
                               keyboardLock);
}


/*============================================================================
 * Function: rsosp_SystemPreset
 * Purpose:  This function sets the instrument parameters to values for good
 *           local/manual interaction. This function corresponds to the
 *           Preset button in the Reset dialog.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_SystemPreset (ViSession instrSession)
{
    return rsosp_sendCmd (instrSession, "SYST:PRES");   
}

/*============================================================================
 * Function: rsosp_GetSystemVersion
 * Purpose:  This function queries the SCPI version with which the instrument
 *           complies.
 *===========================================================================*/
ViStatus _VI_FUNC rsosp_GetSystemVersion (
    ViSession   instrSession,
    ViChar      _VI_FAR SCPIVersion[]
)
{
    ViStatus rsosp_status = VI_SUCCESS;

    rsosp_status = GETPARAMSTRING (instrSession,
                              "SYST:VERS?",
                              SCPIVersion,
                              32);

    return (rsosp_status < VI_SUCCESS) ? rsosp_status : VI_SUCCESS;    
}

/*===========================================================================*/
/* Function: State Checking                                                  */
/* Purpose:  This function switches ON/OFF state checking of the instrument  */
/*           (reading of the Standard Event Register and checking it for     */
/*           error). Nearly all driver function are using state checking.    */
/*           Switch this function to ON when debug your application. For     */
/*           better bus throughput and instruments performance switch it OFF.*/
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_errorCheckState (
    ViSession instrSession,
    ViBoolean state
)
{
    ViStatus            rsosp_status = VI_SUCCESS;
    rsosp_instrRange    instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    CHECKNULL (instrPtr);
    
    instrPtr -> errorChecking = state;

    return rsosp_status;    
}

/*===========================================================================*/
/* Function: Option Checking                                                 */
/* Purpose:  This function switches ON/OFF option checking of the instrument */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_optionCheckState (
    ViSession instrSession,
    ViBoolean state
)
{
    ViStatus            rsosp_status = VI_SUCCESS;
    rsosp_instrRange    instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    instrPtr -> optionChecking = state;

    return rsosp_status;    
}

/*===========================================================================*/
/* Function: Write To Instrument                                             */
/* Purpose:  This function writes a command string to the instrument.        */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_writeInstrData (
    ViSession   instrSession,
    ViString    writeBuffer
)
{
    ViStatus rsosp_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, "%s\n", writeBuffer));

    return rsosp_status;
}

/*===========================================================================*/
/* Function: Read Instrument Buffer                                          */
/* Purpose:  This function reads the output buffer of the instrument.        */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_readInstrData (
    ViSession   instrSession,
    ViInt32     numberBytesToRead,
    ViChar      _VI_FAR readBuffer[],
    ViInt32     *numBytesRead
)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViUInt32    tmpCnt = 0;

    if (numBytesRead != NULL)
        *numBytesRead = 0L;

    CHECKERR (viRead (instrSession, (ViBuf)readBuffer, (ViUInt32) numberBytesToRead, &tmpCnt));

    if (numBytesRead)
        *numBytesRead = (ViInt32)tmpCnt;

    return rsosp_status;    
}

/*===========================================================================*/
/* Function: Reset                                                           */
/* Purpose:  This function resets the instrument.  If the reset function     */
/*           is not supported by the instrument, this function returns       */
/*           the warning VI_WARN_NSUP_RESET.                                 */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_reset (ViSession instrSession)
{
    ViStatus rsosp_status = VI_SUCCESS;

    /*  Initialize the instrument to a known state.  */
    CHECKERR (viPrintf (instrSession, "*RST;*CLS\n"));
    CHECKERR (rsosp_defaultInstrSetup (instrSession));

    return rsosp_status;    
}

/*===========================================================================*/
/* Function: Self-Test                                                       */
/* Purpose:  This function executes the instrument self-test and returns     */
/*           the result. If the self test function is not supported by the   */
/*           instrument, this function returns the warning                   */
/*           VI_WARN_NSUP_SELF_TEST.                                         */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_self_test (
    ViSession   instrSession,
    ViInt16     *testResult,
    ViChar      _VI_FAR testMessage[]
)
{
    ViStatus rsosp_status = VI_SUCCESS;

    CHECKERR( viQueryf (instrSession, "*TST?\n", "%hd", testResult));

    sprintf (testMessage, "%s", (*testResult == 0) ? "Self-Test Passed" : "Self-Test Failed");

    return rsosp_status;    
}

/*===========================================================================*/
/* Function: Error Query                                                     */
/* Purpose:  This function queries the instrument error queue, and returns   */
/*           the result. If the error query function is not supported by the */
/*           instrument, this function returns the warning                   */
/*           VI_WARN_NSUP_ERROR_QUERY.                                       */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_error_query (
    ViSession   instrSession,
    ViInt32     *errorCode, 
    ViChar      _VI_FAR errorMessage[]
)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar      buffer[BUFFER_SIZE] = "";
	ViInt32     oldCode = 0;
	ViChar      oldMessage[256] = "";
	ViChar      *p2buf = NULL;
	ViInt32     i = 0;
	ViUInt32    retCnt = 0;

	p2buf = (char *)malloc(1);

	*p2buf = '\0';

	do
	{
		CHECKERR(viWrite(instrSession, (ViBuf) ":SYST:ERR?\n", 11, VI_NULL));
		CHECKERR(viRead(instrSession, (ViBuf)buffer, BUFFER_SIZE, &retCnt));

		if (sscanf(buffer, "%ld,%*[^0-9a-zA-Z]%256[^\"]", &oldCode, oldMessage) != 2)
		{
			return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
		}
		else
		{
			if (oldCode != 0)
			{
				*errorCode = oldCode;
				p2buf = (char *)realloc(p2buf, strlen(p2buf) + strlen(oldMessage) + 2);
				strcat(p2buf, oldMessage);
				strcat(p2buf, "\n");
				i++;
			}
			else if (i == 0)
			{
				*errorCode = oldCode;
				strncpy(errorMessage, oldMessage, 1024);
			}
		}

	} while (oldCode != 0);

	if (i > 0)
	{
		p2buf[strlen(p2buf) - 1] = '\0';
		strncpy(errorMessage, p2buf, 1024);
	}

	if (p2buf) free(p2buf);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Error Message                                                   */
/* Purpose:  This function translates the error return value from the        */
/*           instrument driver into a user-readable string.                  */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_error_message (
    ViSession   instrSession,
    ViStatus    statusCode, 
    ViChar      _VI_FAR errMessage[]
)           
{
    ViStatus rsosp_status = VI_SUCCESS;
    ViInt16 i;

    static rsosp_tStringValPair statusDescArray[] = {
        {VI_WARN_NSUP_ID_QUERY,                 "WARNING: ID Query not supported"},
        {VI_WARN_NSUP_RESET,                    "WARNING: Reset not supported"},
        {VI_WARN_NSUP_SELF_TEST,                "WARNING: Self-test not supported"},
        {VI_WARN_NSUP_ERROR_QUERY,              "WARNING: Error Query not supported"},
        {VI_WARN_NSUP_REV_QUERY,                "WARNING: Revision Query not supported"},
        {VI_ERROR_FAIL_ID_QUERY,                "ERROR: Identification query failed"},
        {VI_ERROR_INV_RESPONSE,                 "ERROR: Interpreting instrument response"},
        {VI_ERROR_INSTR_FILE_OPEN,              "ERROR: Opening the specified file"},
        {VI_ERROR_INSTR_FILE_WRITE,             "ERROR: Writing to the specified file"},
        {VI_ERROR_INSTR_INTERPRETING_RESPONSE,  "ERROR: Interpreting the instrument's response"},

        /* --- Instrument specific error codes --- */

        {RSOSP_ERROR_INSTRUMENT_ERROR,          "ERROR: Instrument status error"},
    
        {RSOSP_ERROR_INVALID_CONFIGURATION,     "ERROR: Instrument configuration error"},

        /* --- Instrument specific warnings --- */

        {RSOSP_WARNING_SMALL_BUFFER,            "WARNING: Buffer too small"},

        {VI_NULL, VI_NULL}
    };

    rsosp_status = viStatusDesc (instrSession, statusCode, errMessage);
	if (statusCode == VI_ERROR_INSTR_PARAMETER)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		sprintf(errMessage, "ERROR: Invalid parameter '%s'", instrPtr->parameterErrorMessage);
	}
	else
	{
		rsosp_status = viStatusDesc(instrSession, statusCode, errMessage);
		if ((rsosp_status != VI_SUCCESS) || (strstr(errMessage, "Unknown status value") != NULL))
		{
			for (i = 0; statusDescArray[i].stringName; i++)
			{
				if (statusDescArray[i].stringVal == statusCode)
				{
					sprintf(errMessage, "%s", statusDescArray[i].stringName);
					return (VI_SUCCESS);
				}
			}
			sprintf(errMessage, "Unknown Error 0x%X", statusCode);
			return (VI_WARN_UNKNOWN_STATUS);
		}
	}

    rsosp_status = VI_SUCCESS;
    return rsosp_status;
}

/*===========================================================================*/
/* Function: Revision Query                                                  */
/* Purpose:  This function returns the driver and instrument revisions.      */
/*           If the revision query function is not supported by the          */
/*           instrument, this function returns the warning                   */
/*           VI_WARN_NSUP_REV_QUERY.                                         */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_revision_query (
    ViSession   instrSession,
    ViChar      _VI_FAR driverRev[],
    ViChar      _VI_FAR instrRev[]
)
{
    ViStatus rsosp_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, "*IDN?\n"));
    CHECKERR (viScanf (instrSession, "%*[^,],%*[^,],%*[^,],%[^\"\r\n]", instrRev));

    sprintf (driverRev, "%s", RSOSP_REVISION);

    return rsosp_status;    
}

/*===========================================================================*/
/* Function: Close                                                           */
/* Purpose:  This function closes the instrument.                            */
/*===========================================================================*/
ViStatus _VI_FUNC rsosp_close (ViSession instrSession)
{
    rsosp_instrRange    instrPtr = NULL;
    ViSession           rmSession = 0;
    ViStatus            rsosp_status = VI_SUCCESS;

    CHECKERR( viGetAttribute (instrSession, VI_ATTR_RM_SESSION, &rmSession));
    CHECKERR( viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    if (instrPtr != NULL)
        free (instrPtr);

    rsosp_status = viClose (instrSession);
    (void) viClose (rmSession);

    return rsosp_status;    
}

/*===========================================================================*/
/*----------- INSERT INSTRUMENT-DEPENDENT UTILITY ROUTINES HERE -------------*/
/*===========================================================================*/

/*===========================================================================*/
/* Function: Default Instrument Setup                                        */
/* Purpose:  This function sends a default setup to the instrument.  This    */
/*           function is called by the rsosp_reset operation and by the      */
/*           rsosp_init function if the reset option has not been            */
/*           selected.  This function is useful for configuring any          */
/*           instrument settings that are required by the rest of the        */
/*           instrument driver functions such as turning headers ON or OFF   */
/*           or using the long or short form for commands, queries, and data.*/
/*===========================================================================*/
ViStatus rsosp_defaultInstrSetup (ViSession instrSession)
{
    ViChar              buffer[BUFFER_SIZE] = "";
    ViChar              *pTok;
    ViInt32             module;
    ViInt32             instrument;
    ViInt32             port;
    
    ViStatus            rsosp_status = VI_SUCCESS;
    rsosp_instrRange    instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
    
    CHECKNULL (instrPtr);
    
    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    instrPtr -> errorChecking = VI_TRUE;
    instrPtr -> optionChecking = VI_TRUE;
    strcpy (instrPtr -> instrDriverRevision, RSOSP_REVISION);

    /* Check instrument ports and find modules */
    CHECKERR(rsosp_OSPCatalog(instrSession, BUFFER_SIZE, buffer));

    for (pTok = strtok(buffer, ","); pTok != VI_NULL; pTok = strtok(VI_NULL, ","))
    {
        sscanf (pTok, "\"%*[^-]-B%d-F%2dA1%d\"", &module, &instrument, &port);
        
        switch (port)
        {
            case RSOSP_PORT_A11: instrPtr->instrPorts[instrument].A11 = module;
                                break;
            case RSOSP_PORT_A12: instrPtr->instrPorts[instrument].A12 = module;
                                break;
            case RSOSP_PORT_A13: instrPtr->instrPorts[instrument].A13 = module;
                                break;
        }
        
    }
    
    return rsosp_status;
}

/*===========================================================================*/
/* Function: Initialize Clean Up                                             */
/* Purpose:  This function is used only by the rsosp_init function.  When    */
/*           an error is detected this function is called to close the       */
/*           open resource manager and instrument object sessions and to     */
/*           set the instrSession that is returned from rsosp_init to        */
/*           VI_NULL.                                                        */
/*===========================================================================*/
ViStatus rsosp_initCleanUp (
    ViSession   openRMSession,
    ViPSession  openInstrSession,
    ViStatus    currentStatus
)
{
    rsosp_instrRange instrPtr = NULL;

    if (viGetAttribute (*openInstrSession, VI_ATTR_USER_DATA, &instrPtr) >= 0)
        if (instrPtr != NULL)
            free (instrPtr);

    (void) viClose (*openInstrSession);
    (void) viClose (openRMSession);
    *openInstrSession = VI_NULL;

    return currentStatus;
}