/*===========================================================================*/
/*= UTILITY ROUTINES (Non-Exportable Functions) =============================*/
/*===========================================================================*/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "rsosp.h"
#include "rsosp_utility.h"

#define IEEE_ERROR_BITS         0x3C    /* AND mask for ESR checking */
#define BUFFER_SIZE             2048UL    /* File I/O buffer size */

/*===========================================================================*/
/* Function: Long Signed Integer Value Out Of Range - ViInt32                */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean rsosp_invalidViInt32Range(ViSession instrSession, ViInt32 val, ViInt32 min, ViInt32 max)
{
	rsosp_instrRange    instrPtr = NULL;

	viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr);

	/* Error Checking Enabled? */
	if (instrPtr != NULL)
	{
		if (instrPtr->rangeChecking != VI_FALSE)
		{
			return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
		}
	}
	return VI_FALSE;
}

/*===========================================================================*/
/* Function: Long Unsigned Integer Value Out Of Range - ViUInt32             */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean rsosp_invalidViUInt32Range(ViSession instrSession, ViUInt32 val, ViUInt32 min, ViUInt32 max)
{
	rsosp_instrRange    instrPtr = NULL;

	viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr);

	/* Error Checking Enabled? */
	if (instrPtr != NULL)
	{
		if (instrPtr->rangeChecking != VI_FALSE)
		{
			(ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
		}
	}
	return VI_FALSE;
}

/*===========================================================================*/
/* Function: Real (Double) Value Out Of Range - ViReal64                     */
/* Purpose:  This function checks a real (double) value to see if it lies    */
/*           between a minimum and maximum value.  If the value is out of    */
/*           range, the return value is VI_TRUE, otherwise the return value  */
/*           is VI_FALSE.                                                    */
/*===========================================================================*/
ViBoolean rsosp_invalidViReal64Range(ViSession instrSession, ViReal64 val, ViReal64 min, ViReal64 max)
{
	rsosp_instrRange    instrPtr = NULL;

	viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr);

	/* Error Checking Enabled? */
	if (instrPtr != NULL)
	{
		if (instrPtr->rangeChecking != VI_FALSE)
		{
			return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
		}
	}
	return VI_FALSE;
}
/*=========================================================================*/
/* Function: Check Status                                                  */
/* Purpose:  This function reads Event Status Register to check for        */
/*           possible errors.                                              */
/*=========================================================================*/
ViStatus rsosp_checkStatus (ViSession instrSession)
{
    ViStatus            rsosp_status = VI_SUCCESS;
    rsosp_instrRange    instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    /* Error Checking Enabled? */
    if (instrPtr != NULL)
    {
        if (instrPtr -> errorChecking != VI_FALSE)
        {
            ViInt32 esrByte = 0;

            /*--------------- Read the Standard Event Register ------------------*/
            CHECKERR (viQueryf (instrSession, "*ESR?\n", "%ld", &esrByte));

            /*---------------- Check if any error or message bit was asserted ---*/
            rsosp_status = ((esrByte & IEEE_ERROR_BITS) != 0) ? RSOSP_ERROR_INSTRUMENT_ERROR : VI_SUCCESS;
        }
    }

    return rsosp_status;
}
/*=========================================================================*/
/* Function: Check Option                                                  */
/* Purpose:  This function reads and compares installed modules with 	   */
/*			 required module.										       */
/*=========================================================================*/
ViStatus rsosp_checkOption(ViSession instrSession, ViString channelList, ViInt32 option)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViInt32 	instrument;
	ViInt32		port;
	ViChar		*pTok;
	ViChar		address[BUFFER_SIZE];
	ViChar  	*pChannelList = channelList;
    rsosp_instrRange    instrPtr = NULL; \

	CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
    if (instrPtr == NULL)
        return rsosp_ParameterError (instrSession, "instrSession");

	if (instrPtr->optionChecking == VI_TRUE)
	{
		do
		{
			pTok = strchr(pChannelList, ','); 		
			if (pTok)
			{
				strncpy(address, pChannelList, pTok - pChannelList);
				address[pTok - pChannelList] = '\0';
				pChannelList = ++pTok;
			}
			else
			{
				strncpy(address, pChannelList, strlen(pChannelList));	
				address[strlen(pChannelList)] = '\0';
			}
		
			RemoveSurroundingWhiteSpace(address);		
			sscanf(address, "F%dA1%d", &instrument, &port);
			switch(port)
			{
				case RSOSP_PORT_A11: {
							if (instrPtr->instrPorts[instrument].A11 != option)	
								rsosp_status = RSOSP_ERROR_INVALID_CONFIGURATION;
						} break;
				case RSOSP_PORT_A12: {
							if (instrPtr->instrPorts[instrument].A12 != option)	
								rsosp_status = RSOSP_ERROR_INVALID_CONFIGURATION;
						} break;
				case RSOSP_PORT_A13: {
							if (instrPtr->instrPorts[instrument].A13 != option)	
								rsosp_status = RSOSP_ERROR_INVALID_CONFIGURATION;\
						} break;
			}
		} while(pTok != VI_NULL);
	}
	
	return rsosp_status;
}
/*===========================================================================*/
/* Function: Send Command                                                    */
/* Purpose:  This function sends command to the instrument with error check. */
/*===========================================================================*/
ViStatus rsosp_sendCmd (ViSession instrSession, ViString cmd)
{
    ViStatus    rsosp_status    = VI_SUCCESS;
    size_t      length          = 0;
    ViChar      *pbuffer;

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc(length + 1);
    if (pbuffer == NULL)
        return VI_ERROR_ALLOC;

    memcpy (pbuffer, cmd, length);
    pbuffer[length] = '\n';

    length++;
    GOCHECKERR (viWrite (instrSession, (ViPBuf) cmd, (ViUInt32) length, VI_NULL));
    GOCHECKERR (rsosp_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rsosp_status;
}

/*===========================================================================*/
/* Function: Set ViInt32                                                     */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsosp_setViInt32(ViSession instrSession,
	ViString cmd,
	ViInt32 value,
	ViInt32 min,
	ViInt32 max,
	ViString parameterName)
{
	ViChar      *wrt_buf;
	ViStatus    rsosp_status = VI_SUCCESS;

	if ((max - min) > 0)
		if (rsosp_invalidViInt32Range(instrSession, value, min, max) == VI_TRUE)
		{
			rsosp_instrRange   instrPtr;
			CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
			return VI_ERROR_INSTR_PARAMETER;
		}


	wrt_buf = (ViChar*)malloc(strlen(cmd) + 20);

	sprintf(wrt_buf, "%s %ld", cmd, value);

	rsosp_status = rsosp_sendCmd(instrSession, wrt_buf);

	free(wrt_buf);
	return rsosp_status;
}

/*===========================================================================*/
/* Function: Get ViInt32                                                     */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsosp_getViInt32(ViSession instrSession,
	ViString cmd,
	ViInt32 *value,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	CHECKERR(viPrintf(instrSession, "%s\n", cmd));
	CHECKERR(viScanf(instrSession, "%ld", value));
	CHECKERR(rsosp_checkStatus(instrSession));

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Get ViUInt32                                                    */
/* Purpose:  This function gets command with unsigned integer parameter.     */
/*===========================================================================*/
ViStatus rsosp_getViUInt32(ViSession instrSession,
	ViString cmd,
	ViUInt32 *value,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	CHECKERR(viPrintf(instrSession, "%s\n", cmd));
	CHECKERR(viScanf(instrSession, "%u", value));
	CHECKERR(rsosp_checkStatus(instrSession));

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Set ViBoolean                                                   */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsosp_setViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean value)
{
    return rsosp_setViString (instrSession,
                              cmd,
                              (value == VI_TRUE) ? "ON" : "OFF",
                              0);
}

/*===========================================================================*/
/* Function: Get ViBoolean                                                   */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsosp_getViBoolean(ViSession instrSession,
	ViString cmd,
	ViBoolean *value,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar      buffer[4];

	if (value == NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	CHECKERR(rsosp_getViString(instrSession,
		cmd,
		buffer, 4,
		0));

	*value = (ViBoolean)((strncmp(buffer, "1", 1) == 0) || (strncmp(buffer, "ON", 2) == 0));

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Set ViReal64                                                    */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rsosp_setViReal64(ViSession instrSession,
	ViString cmd,
	ViReal64 value,
	ViReal64 min,
	ViReal64 max,
	ViString parameterName)
{
	ViChar      buffer[BUFFER_SIZE];
	ViStatus    rsosp_status = VI_SUCCESS;

	/* do range checking only if there's difference between max and min */
	if ((max - min) > 1.0e-14)
		if (rsosp_invalidViReal64Range(instrSession, value, min, max) == VI_TRUE)
		{
			rsosp_instrRange   instrPtr;
			CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
			return VI_ERROR_INSTR_PARAMETER;
		}

	sprintf(buffer, "%s %.12lG", cmd, value);

	return rsosp_sendCmd(instrSession, buffer);
}

/*===========================================================================*/
/* Function: Get ViReal64                                                    */
/* Purpose:  This function gets command with double parameter.               */
/*===========================================================================*/
ViStatus rsosp_getViReal64(ViSession instrSession,
	ViString cmd,
	ViReal64 *value,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	CHECKERR(viPrintf(instrSession, "%s\n", cmd));
	CHECKERR(viScanf(instrSession, "%le", value));
	CHECKERR(rsosp_checkStatus(instrSession));

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Set ViReal64 with Units                                         */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rsosp_setViReal64Units(ViSession instrSession,
	ViString cmd,
	ViReal64 value,
	ViString units,
	ViReal64 min,
	ViReal64 max,
	ViString parameterName)
{
	ViChar      buffer[BUFFER_SIZE];
	ViStatus    rsosp_status = VI_SUCCESS;

	/* do range checking only if there's difference between max and min */
	if ((max - min) > 1.0e-14)
		if (rsosp_invalidViReal64Range(instrSession, value, min, max) == VI_TRUE)
		{
			rsosp_instrRange   instrPtr;
			CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
			return VI_ERROR_INSTR_PARAMETER;
		}

	sprintf(buffer, "%s %.12f %s", cmd, value, units);

	return rsosp_sendCmd(instrSession, buffer);
}

/*===========================================================================*/
/* Function: Set String                                                      */
/* Purpose:  This function sets command with string parameter.               */
/*===========================================================================*/
ViStatus rsosp_setViString(ViSession instrSession,
	ViString cmd,
	ViString value,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar*     wrt_buf;
	ViChar*     p2buf;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	wrt_buf = (ViChar*)malloc(strlen(cmd) + strlen(value) + 4);
	p2buf = wrt_buf;
	p2buf += sprintf(wrt_buf, "%s %s\n", cmd, value);

	GOCHECKERR(viWrite(instrSession, (ViBuf)wrt_buf, (ViUInt32)(p2buf - wrt_buf), VI_NULL));
	GOCHECKERR(rsosp_checkStatus(instrSession));

ERROR_LABEL:

	free(wrt_buf);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Get String                                                      */
/* Purpose:  This function gets command with string parameter.               */
/*===========================================================================*/
ViStatus rsosp_getViString(ViSession instrSession,
	ViString cmd,
	ViString value,
	ViUInt32 buffer_size,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar      *pbuffer,
		*p2buffer;
	ViUInt32    count = 0;
	size_t      length = 0;

	if (value == NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	length = strlen(cmd);
	pbuffer = (ViChar*)malloc(length + 1);

	memcpy(pbuffer, cmd, length);
	if (pbuffer == NULL)
		return VI_ERROR_ALLOC;

	pbuffer[length] = '\n';

	GOCHECKERR(viWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)(length + 1), VI_NULL));

	free(pbuffer);

	pbuffer = NULL;

	GOCHECKERR(rsosp_readDataUnknownLength(instrSession, &pbuffer, &count));

	if (pbuffer == NULL)
		return rsosp_status;

	GOCHECKERR(rsosp_checkStatus(instrSession));

	if (count > buffer_size)
	{
		rsosp_status = RSOSP_WARNING_SMALL_BUFFER;
		count = buffer_size;
		memcpy(value, pbuffer, (size_t)count);
		value[count] = '\0';
	}
	else
		if ((p2buffer = strpbrk(pbuffer, "\r\n")) != VI_NULL)
		{
			rsosp_status = VI_SUCCESS;
			memcpy(value, pbuffer, (size_t)(p2buffer - pbuffer));
			value[p2buffer - pbuffer] = '\0';
		}

ERROR_LABEL:

	if (pbuffer != NULL)
		free(pbuffer);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Set Quoted String                                               */
/* Purpose:  This function sets command with quoted string parameter.        */
/*===========================================================================*/
ViStatus rsosp_setViStringQuoted(ViSession instrSession,
	ViString cmd,
	ViString quotes,
	ViString value,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar*     wrt_buf;
	ViChar*     p2buf;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	wrt_buf = (ViChar*)malloc(strlen(cmd) + strlen(value) + 2 * strlen(quotes) + 4);
	p2buf = wrt_buf;
	p2buf += sprintf(wrt_buf, "%s %s%s%s\n", cmd, quotes, value, quotes);

	GOCHECKERR(viWrite(instrSession, (ViBuf)wrt_buf, (ViUInt32)(p2buf - wrt_buf), VI_NULL));
	GOCHECKERR(rsosp_checkStatus(instrSession));

ERROR_LABEL:

	free(wrt_buf);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Set Enum                                                        */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rsosp_setEnum(ViSession instrSession,
	ViString cmd,
	ViInt32 value,
	ViString list[],
	ViString parameterName)
{
	ViStatus rsosp_status = VI_SUCCESS;

	CHECKERR(rsosp_checkIndexInList(instrSession, value, list, parameterName));

	return rsosp_setViString(instrSession, cmd, list[value], parameterName);
}

/*===========================================================================*/
/* Function: Get Enum                                                        */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rsosp_getEnum(ViSession instrSession,
	ViString cmd,
	ViInt32 *value,
	ViString list[],
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar      buffer[BUFFER_SIZE] = "";

	CHECKERR(rsosp_getViString(instrSession, cmd, buffer, BUFFER_SIZE, parameterName));
	CHECKERR(rsosp_getIndexFromString(buffer, list, value));

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Check Index List                                                */
/* Purpose:  This function checks if the passed index is valid for given list*/
/*===========================================================================*/
ViStatus rsosp_checkIndexInList(ViSession instrSession,
	ViInt32 index,
	ViString list[],
	ViString parameterName)

{
	ViInt32     i = 0;
	ViStatus    rsosp_status = VI_SUCCESS;

	if (index < 0)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	while (list[i++] != VI_NULL);
	if (index >= i)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	return VI_SUCCESS;
}

/*===========================================================================*/
/* Function: Get Index From String                                           */
/* Purpose:  This function returns index for string.                         */
/*===========================================================================*/
ViStatus rsosp_getIndexFromString (ViString string,
                                   ViString *list,
                                   ViInt32 *index)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViUInt32    i;
    size_t      len;

    for (i = 0; list[i]; i++)
        if ((len = strlen (string)) == strlen (list[i])) {
            if (strncmp(string, list[i], strlen (string)) == 0) break;
        }

    *index = (ViInt32) i;

    return (list[i]) ? rsosp_status : VI_ERROR_INSTR_INTERPRETING_RESPONSE;
}

/*===========================================================================*/
/* Function: Get Catalog                                                     */
/*===========================================================================*/
ViStatus rsosp_getCatalog(ViSession instrSession,
	ViString cmd,
	ViUInt32 arraySize,
	ViChar* array,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar      *pbuffer,
	*p2buffer;
	size_t      count = 0,
	length = 0;
	
	if (array == NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	length = strlen(cmd);
	pbuffer = (ViChar*)malloc(length + 1);

	if (pbuffer == NULL)
		return VI_ERROR_ALLOC;

	memcpy(pbuffer, cmd, length);
	pbuffer[length] = '\n';

	GOCHECKERR(viWrite(instrSession, (ViPBuf)pbuffer, (ViUInt32)(length + 1), VI_NULL));

	free(pbuffer);

	pbuffer = NULL;

	GOCHECKERR(rsosp_readDataUnknownLength(instrSession, &pbuffer, &count));
	GOCHECKERR(rsosp_checkStatus(instrSession));

	if (count > (size_t)arraySize)
	{
		rsosp_status = RSOSP_WARNING_SMALL_BUFFER;
		count = (size_t)arraySize;
		memcpy(array, pbuffer, count);
		array[count - 1] = '\0';
	}
	else
	{
		if ((p2buffer = strchr(pbuffer, ';')) != VI_NULL)
		{
			rsosp_status = VI_SUCCESS;
			memcpy(array, pbuffer, (size_t)(p2buffer - pbuffer));
			array[p2buffer - pbuffer] = '\0';
		}
		else
		{
			memcpy(array, pbuffer, count);
			array[count - 1] = '\0';
		}
	}
ERROR_LABEL:

	if (pbuffer != NULL)
		free(pbuffer);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Get ViReal64 Array                                              */
/*===========================================================================*/
ViStatus rsosp_getViReal64Array(ViSession instrSession,
	ViString cmd,
	ViReal64 *value,
	ViInt32  count,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar*     pbuffer = VI_NULL;
	ViChar*     p2buf;
	ViUInt32    local_count = 0;
	ViInt32     i = 0;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	GOCHECKERR(viPrintf(instrSession, "%s\n", cmd));
	GOCHECKERR(rsosp_readDataUnknownLength(instrSession, &pbuffer, &local_count));
	p2buf = pbuffer;

	p2buf = strtok(pbuffer, ",");

	while ((p2buf != NULL) && (i < count))
	{
		value[i++] = atof(p2buf);
		p2buf = strtok(NULL, ",");
	};

	GOCHECKERR(rsosp_checkStatus(instrSession));

ERROR_LABEL:

	free(pbuffer);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Get ViInt32 Array                                              */
/*===========================================================================*/
ViStatus rsosp_getViInt32Array(ViSession instrSession,
	ViString cmd,
	ViInt32 *value,
	ViInt32  count,
	ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	ViChar*     pbuffer = VI_NULL;
	ViChar*     p2buf;
	ViUInt32    local_count = 0;
	ViInt32     i = 0;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

	GOCHECKERR(viPrintf(instrSession, "%s\n", cmd));
	GOCHECKERR(rsosp_readDataUnknownLength(instrSession, &pbuffer, &local_count));
	p2buf = pbuffer;

	p2buf = strtok(pbuffer, ",");

	while ((p2buf != NULL) && (i < count))
	{
		value[i++] = atol(p2buf);
		p2buf = strtok(NULL, ",");
	};

	GOCHECKERR(rsosp_checkStatus(instrSession));

ERROR_LABEL:

	free(pbuffer);

	return rsosp_status;
}

/*===========================================================================*/
/* Function: Get ViUInt32 Array                                              */
/*===========================================================================*/
ViStatus rsosp_getViUInt32Array (ViSession instrSession,
                             ViString cmd,
                             ViUInt32 *value,
                             ViInt32  count,
							 ViString parameterName)
{
    ViStatus    rsosp_status = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViChar*     p2buf;
    ViUInt32    local_count = 0;
    ViInt32     i = 0;

	if (value == VI_NULL)
	{
		rsosp_instrRange   instrPtr;
		CHECKERR(viGetAttribute(instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
		return VI_ERROR_INSTR_PARAMETER;
	}

    GOCHECKERR (viPrintf(instrSession, "%s\n",cmd));
    GOCHECKERR (rsosp_readDataUnknownLength(instrSession, &pbuffer, &local_count));
    p2buf = pbuffer;

    p2buf = strtok(pbuffer, ",");

    while ((p2buf != NULL) && (i < count))
    {
        value[i++] = strtoul(p2buf, NULL, 10);
        p2buf = strtok(NULL, ",");
    };

    GOCHECKERR (rsosp_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rsosp_status;
}

/*===========================================================================*/
/* Function: Read Data With Unknown length                                   */
/* Purpose:  This function allocates buffer and reads data from the          */
/*           instrument                                                      */
/* WARNING:  You must deallocate buffer, if any pointer has been returned.   */
/*           This function does not work correctly, when you read bin data   */
/*           via serial intf                                                 */
/*===========================================================================*/
ViStatus rsosp_readDataUnknownLength(ViSession instrSession,
                                     /*@null@*/ViChar **outBuffer,
                                     ViUInt32 *count)
{
    ViStatus rsosp_status = VI_SUCCESS;
    ViUInt32 actSize, retCnt = 0;
    ViChar   *buffer = NULL, *p2buffer = NULL;

    *outBuffer = VI_NULL;
    if (count) *count = 0;

    /* Allocate Buffer */
    actSize = BUFFER_SIZE;
    buffer = (ViChar *)calloc(actSize, (size_t) sizeof(ViChar));
    if (buffer == VI_NULL) return VI_ERROR_ALLOC;

    p2buffer = buffer;
    do
    {
        if ((rsosp_status = viRead (instrSession, (ViPBuf)p2buffer, BUFFER_SIZE, &retCnt)) < 0)
        {
            free(buffer);
            *outBuffer = VI_NULL;
            return rsosp_status;
        }
        if (retCnt == BUFFER_SIZE) /* Buffer was small, reallocate it */
        {
            actSize += BUFFER_SIZE;
            buffer = (ViChar *)realloc(buffer, (size_t) actSize);
            if (buffer == VI_NULL) return VI_ERROR_ALLOC;
            p2buffer = buffer + (actSize - BUFFER_SIZE); /* Set pointer to end of data in rallocated buffer */
        }
        else
        {
            *(p2buffer+retCnt) = '\0'; /* For sure set end of string at the end of data */
        }
    } while (rsosp_status == VI_SUCCESS_MAX_CNT); /* if buffer was small, continue with next read */

    *outBuffer = buffer;
    if (count != NULL) *count = (actSize - BUFFER_SIZE) + retCnt;

    return rsosp_status;
}

/*===========================================================================*/
/* Function: RemoveSurroundingWhiteSpace									 */
/* Purpose:  This function removes surrounding white space around a string   */
/*===========================================================================*/
void RemoveSurroundingWhiteSpace(ViChar string[])
{
    ViChar  *start = string,
            *end;
    ViInt32 lengthWithoutLeadingWS;

    if (start)
        {
        while (*start && isspace((unsigned char)*start))
            start++;

        lengthWithoutLeadingWS = (ViInt32)strlen( start);
        memmove( string, start, lengthWithoutLeadingWS);
        end = string + lengthWithoutLeadingWS;
        while (end > string && isspace((unsigned char)*(end-1)))
            end--;

        *end = '\0';
        }
}

ViStatus rsosp_ParameterError(ViSession io, ViString parameterName)
{
	ViStatus    rsosp_status = VI_SUCCESS;
	rsosp_instrRange   instrPtr;
	CHECKERR(viGetAttribute(io, VI_ATTR_USER_DATA, &instrPtr));
	strncpy(instrPtr->parameterErrorMessage, parameterName, 255);
	return VI_ERROR_INSTR_PARAMETER;
}

